
import enum

from minikerberos.gssapi.gssapi import get_gssapi, GSSWrapToken
from minikerberos.protocol.asn1_structs import AP_REQ, AP_REP, TGS_REP
from minikerberos.protocol.encryption import Enctype, Key, _enctype_table
from wsnet.clientauth import WSNETAuth
from asyauth.common.winapi.constants import ISC_REQ

import enum
import io
import os

from asn1crypto.core import ObjectIdentifier

class KRB5_MECH_INDEP_TOKEN:
	# https://tools.ietf.org/html/rfc2743#page-81
	# Mechanism-Independent Token Format

	def __init__(self, data, oid, remlen = None):
		self.oid = oid
		self.data = data

		#dont set this
		self.length = remlen
	
	@staticmethod
	def from_bytes(data):
		return KRB5_MECH_INDEP_TOKEN.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		
		start = buff.read(1)
		if start != b'\x60':
			raise Exception('Incorrect token data!')
		remaining_length = KRB5_MECH_INDEP_TOKEN.decode_length_buffer(buff)
		token_data = buff.read(remaining_length)
		
		buff = io.BytesIO(token_data)
		pos = buff.tell()
		buff.read(1)
		oid_length = KRB5_MECH_INDEP_TOKEN.decode_length_buffer(buff)
		buff.seek(pos)
		token_oid = ObjectIdentifier.load(buff.read(oid_length+2))
		
		return KRB5_MECH_INDEP_TOKEN(buff.read(), str(token_oid), remlen = remaining_length)
		
	@staticmethod
	def decode_length_buffer(buff):
		lf = buff.read(1)[0]
		if lf <= 127:
			length = lf
		else:
			bcount = lf - 128
			length = int.from_bytes(buff.read(bcount), byteorder = 'big', signed = False)
		return length
		
	@staticmethod
	def encode_length(length):
		if length <= 127:
			return length.to_bytes(1, byteorder = 'big', signed = False)
		else:
			lb = length.to_bytes((length.bit_length() + 7) // 8, 'big')
			return (128+len(lb)).to_bytes(1, byteorder = 'big', signed = False) + lb
		
		
	def to_bytes(self):
		t = ObjectIdentifier(self.oid).dump() + self.data
		t = b'\x60' + KRB5_MECH_INDEP_TOKEN.encode_length(len(t)) + t
		return t[:-len(self.data)] , self.data

class KerberosClientWSNET:
	def __init__(self, settings):
		self.iterations = 0
		self.settings = settings
		self.ksspi = WSNETAuth()
		self.client = None
		self.target = None
		self.gssapi = None
		self.etype = None
		self.session_key = None
		self.seq_number = None
		
		self.setup()
		
	def setup(self):
		return

	def get_seq_number(self):
		return self.seq_number
		
	async def encrypt(self, data, message_no):
		return self.gssapi.GSS_Wrap(data, message_no)
		
	async def decrypt(self, data, message_no, direction='init', auth_data=None):
		return self.gssapi.GSS_Unwrap(data, message_no, direction=direction, auth_data=auth_data)
	
	def get_session_key(self):
		return self.session_key
	
	async def authenticate(self, authData = None, flags = ISC_REQ.CONNECTION, seq_number = 0, is_rpc = False):
		try:
			if is_rpc == True:
				if self.iterations == 0:
					flags = ISC_REQ.CONFIDENTIALITY | \
							ISC_REQ.INTEGRITY | \
							ISC_REQ.MUTUAL_AUTH | \
							ISC_REQ.REPLAY_DETECT | \
							ISC_REQ.SEQUENCE_DETECT|\
							ISC_REQ.USE_DCE_STYLE
					
					
					status, ctxattr, apreq, err = await self.ksspi.authenticate('KERBEROS', '', 'termsrv/%s' % self.settings.target, 3, flags.value, authdata = b'')
					if err is not None:
						raise err
					self.iterations += 1
					return apreq, True, None
				
				elif self.iterations == 1:
					status, ctxattr, data, err = await self.ksspi.authenticate('KERBEROS', '','termsrv/%s' % self.settings.target, 3, flags.value, authdata = authData)
					if err is not None:
						return None, None, err
					self.session_key, err = await self.ksspi.get_sessionkey()
					if err is not None:
						return None, None, err
						
					aprep = AP_REP.load(data).native
					subkey = Key(aprep['enc-part']['etype'], self.session_key)
					self.gssapi = get_gssapi(subkey)

					if aprep['enc-part']['etype'] != 23: #no need for seq number in rc4
						raw_seq_data, err = await self.ksspi.get_sequenceno()
						if err is not None:
							return None, None, err
						self.seq_number = GSSWrapToken.from_bytes(raw_seq_data[16:]).SND_SEQ
					
					self.iterations += 1
					await self.ksspi.disconnect()
					return data, False, None
					
				else:
					raise Exception('SSPI Kerberos -RPC - auth encountered too many calls for authenticate.')
			
				
			else:
				status, ctxattr, apreq, err = await self.ksspi.authenticate('KERBEROS', '','termsrv/%s' % self.settings.target, 3, flags.value, authdata = b'')
				if err is not None:
					return None, None, err
				
				self.session_key, err = await self.ksspi.get_sessionkey()
				if err is not None:
					raise err
				await self.ksspi.disconnect()

				return apreq, False, None
		except Exception as e:
			return None, None, e
		


import enum
import datetime
import ctypes

from asyauth.common.winapi.constants import SECBUFFER_TYPE, SEC_E, ISC_REQ
from minikerberos.gssapi.gssapi import GSSWrapToken

windll      = ctypes.windll
byref       = ctypes.byref
addressof   = ctypes.addressof
sizeof      = ctypes.sizeof
SIZEOF      = ctypes.sizeof

LPSTR   = ctypes.c_char_p
CHAR    = ctypes.c_char
PCHAR   = LPSTR
DWORD   = ctypes.c_uint32
POINTER = ctypes.POINTER
LONG    = ctypes.c_int32
ULONG   = ctypes.c_uint32
LPVOID  = ctypes.c_void_p
BYTE    = ctypes.c_ubyte
PVOID   = LPVOID
PPVOID  = POINTER(PVOID)
LPBYTE  = POINTER(BYTE)
LPULONG = POINTER(ULONG)
PULONG  = LPULONG

class Structure(ctypes.Structure):
	if sizeof(ctypes.c_void_p) == 4:
		_pack_ = 1


# call API to get max token size, or..
maxtoken_size = 2880 # bytes

_FILETIME_null_date = datetime.datetime(1601, 1, 1, 0, 0, 0)
def FiletimeToDateTime(ft):
	timestamp = (ft.dwHighDateTime << 32) + ft.dwLowDateTime
	return _FILETIME_null_date + datetime.timedelta(microseconds=timestamp/10)

#timestamp is LARGE_INTEGER
#same as FILETIME structure

#https://docs.microsoft.com/en-us/windows/desktop/api/minwinbase/ns-minwinbase-filetime
class FILETIME(Structure):
	_fields_ = [
		("dwLowDateTime",   DWORD),
		("dwHighDateTime",   DWORD),
	]
PFILETIME = POINTER(FILETIME)
TimeStamp = FILETIME
PTimeStamp = PFILETIME

SEC_CHAR = CHAR
PSEC_CHAR = PCHAR

class LUID(Structure):
	_fields_ = [
		("LowPart",	 DWORD),
		("HighPart",	LONG),
	]

PLUID = POINTER(LUID)

class SEC_CHANNEL_BINDINGS(Structure):
	_fields_ = [
		('dwInitiatorAddrType',ULONG),
		('cbInitiatorLength',ULONG),
		('dwInitiatorOffset',ULONG),
		('dwAcceptorAddrType',ULONG),
		('cbAcceptorLength',ULONG),
		('dwAcceptorOffset',ULONG),
		('cbApplicationDataLength',ULONG),
		('dwApplicationDataOffset',ULONG),
		('Buffer', LPVOID)
	]
	
# https://docs.microsoft.com/en-us/windows/desktop/api/sspi/ns-sspi-secpkgcontext_sessionkey
class SecPkgContext_SessionKey(Structure):
	_fields_ = [('SessionKeyLength',ULONG),('SessionKey',LPBYTE)]
	
	@property
	def Buffer(self):
		return ctypes.string_at(self.SessionKey, size=self.SessionKeyLength)

# https://github.com/benjimin/pywebcorp/blob/master/pywebcorp/ctypes_sspi.py
class SecHandle(ctypes.Structure): 
	
	_fields_ = [('dwLower',PULONG),('dwUpper',PULONG)]
	def __init__(self): # populate deeply (empty memory fields) rather than shallow null POINTERs.
		ctypes.Structure.__init__(self, ctypes.pointer(ULONG()), ctypes.pointer(ULONG()))

class SecBuffer(Structure):
	"""Stores a memory buffer: size, type-flag, and POINTER. 
	The type can be empty (0) or token (2).
	InitializeSecurityContext will write to the buffer that is flagged "token"
	and update the size, or else fail 0x80090321=SEC_E_BUFFER_TOO_SMALL."""	
	_fields_ = [('cbBuffer',ULONG),('BufferType',ULONG),('pvBuffer',PVOID)]
	def __init__(self, token=b'\x00'*maxtoken_size, buffer_type = SECBUFFER_TYPE.SECBUFFER_TOKEN):
		buf = ctypes.create_string_buffer(token, size=len(token)) 
		Structure.__init__(self,sizeof(buf),buffer_type.value,ctypes.cast(ctypes.pointer(buf),PVOID))
	@property
	def Buffer(self):
		return (SECBUFFER_TYPE(self.BufferType), ctypes.string_at(self.pvBuffer, size=self.cbBuffer))	 

class SecBufferDesc(Structure):
	"""Descriptor stores SECBUFFER_VERSION=0, number of buffers (e.g. one),
	and POINTER to an array of SecBuffer structs."""
	_fields_ = [('ulVersion',ULONG),('cBuffers',ULONG),('pBuffers',POINTER(SecBuffer))]
	def __init__(self, secbuffers = None):
		#secbuffers = a list of security buffers (SecBuffer)
		if secbuffers is not None:
			Structure.__init__(self,0,len(secbuffers),(SecBuffer * len(secbuffers))(*secbuffers))
		else:
			Structure.__init__(self,0,1,ctypes.pointer(SecBuffer()))
	def __getitem__(self, index):
		return self.pBuffers[index]
		
	@property
	def Buffers(self):
		data = []
		for i in range(self.cBuffers):
			data.append(self.pBuffers[i].Buffer)
		return data
		
PSecBufferDesc = POINTER(SecBufferDesc)
PSecHandle     = POINTER(SecHandle)
CredHandle     = SecHandle
PCredHandle    = PSecHandle
CtxtHandle     = SecHandle
PCtxtHandle    = PSecHandle


class DecryptFlags(enum.Enum):										
	SIGN_ONLY = 0
	SECQOP_WRAP_NO_ENCRYPT = 2147483649 # same as KERB_WRAP_NO_ENCRYPT
	
def FreeContextBuffer(secbuff):
	def errc(result, func, arguments):
		if SEC_E(result) == SEC_E.OK:
			return SEC_E(result)
		raise Exception('%s failed with error code %s (%s)' % ('DecryptMessage', result, SEC_E(result)))
	
	_FreeContextBuffer = windll.Secur32.FreeContextBuffer
	_FreeContextBuffer.argtypes = [PVOID]
	_FreeContextBuffer.restype  = DWORD
	_FreeContextBuffer.errcheck  = errc
	
	res = _FreeContextBuffer(byref(secbuff))
	return

#https://github.com/mhammond/pywin32/blob/d64fac8d7bda2cb1d81e2c9366daf99e802e327f/win32/Lib/sspi.py#L108
#https://docs.microsoft.com/en-us/windows/desktop/secauthn/using-sspi-with-a-windows-sockets-client
#https://msdn.microsoft.com/en-us/library/Aa374712(v=VS.85).aspx
def AcquireCredentialsHandle(client_name, package_name, tragetspn, cred_usage, pluid = None, authdata = None):
	def errc(result, func, arguments):
		if SEC_E(result) == SEC_E.OK:
			return result
		raise Exception('%s failed with error code %s (%s)' % ('AcquireCredentialsHandle', result, SEC_E(result)))
		
	_AcquireCredentialsHandle = windll.Secur32.AcquireCredentialsHandleA
	_AcquireCredentialsHandle.argtypes = [PSEC_CHAR, PSEC_CHAR, ULONG, PLUID, PVOID, PVOID, PVOID, PCredHandle, PTimeStamp]
	_AcquireCredentialsHandle.restype  = DWORD
	_AcquireCredentialsHandle.errcheck  = errc
	
	#TODO: package_name might be different from version to version. implement functionality to poll it properly!
	
	cn = None
	if client_name:
		cn = LPSTR(client_name.encode('ascii'))
	pn = LPSTR(package_name.encode('ascii'))
	
	creds = CredHandle()
	ts = TimeStamp()
	res = _AcquireCredentialsHandle(cn, pn, cred_usage, pluid, authdata, None, None, byref(creds), byref(ts))
	return creds
	
# https://msdn.microsoft.com/en-us/library/windows/desktop/aa375507(v=vs.85).aspx
def InitializeSecurityContext(creds, target, ctx = None, flags = ISC_REQ.INTEGRITY | ISC_REQ.CONFIDENTIALITY | ISC_REQ.SEQUENCE_DETECT | ISC_REQ.REPLAY_DETECT, TargetDataRep  = 0, token = None, cb_data = None):
	def errc(result, func, arguments):
		if SEC_E(result) in [SEC_E.OK, SEC_E.COMPLETE_AND_CONTINUE, SEC_E.COMPLETE_NEEDED, SEC_E.CONTINUE_NEEDED, SEC_E.INCOMPLETE_CREDENTIALS]:
			return SEC_E(result)
		raise Exception('%s failed with error code %s (%s) Target: %s' % ('InitializeSecurityContext', result, SEC_E(result), target))
		
	_InitializeSecurityContext = windll.Secur32.InitializeSecurityContextA
	_InitializeSecurityContext.argtypes = [PCredHandle, PCtxtHandle, PSEC_CHAR, ULONG, ULONG, ULONG, PSecBufferDesc, ULONG, PCtxtHandle, PSecBufferDesc, PULONG, PTimeStamp]
	_InitializeSecurityContext.restype  = DWORD
	_InitializeSecurityContext.errcheck  = errc
	if target:
		ptarget = LPSTR(target.encode('ascii'))
	else:
		ptarget = None
	newbuf = SecBufferDesc()
	outputflags = ULONG()
	expiry = TimeStamp()
	
	sb = []
	if token:
		sb = [SecBuffer(token)]
	if cb_data:
		cb = b'\x00'*24 + len(cb_data).to_bytes(4, byteorder='little', signed=False) + (32).to_bytes(4, byteorder='little', signed=False) + cb_data
		sb.append(SecBuffer(token=cb, buffer_type=SECBUFFER_TYPE.SECBUFFER_CHANNEL_BINDINGS))
	token = SecBufferDesc(sb)

	if not ctx:
		ctx = CtxtHandle()
		res = _InitializeSecurityContext(byref(creds), None, ptarget, int(flags), 0 ,TargetDataRep, byref(token) if token else None, 0, byref(ctx), byref(newbuf), byref(outputflags), byref(expiry))
	else:
		res = _InitializeSecurityContext(byref(creds), byref(ctx), ptarget, int(flags), 0 ,TargetDataRep, byref(token) if token else None, 0, byref(ctx), byref(newbuf), byref(outputflags), byref(expiry))
	
	data = newbuf.Buffers
	return res, ctx, data[0][1], ISC_REQ(outputflags.value), expiry
	
def DecryptMessage(ctx, token, data, message_no = 0):
	def errc(result, func, arguments):
		if SEC_E(result) == SEC_E.OK:
			return SEC_E(result)
		raise Exception('%s failed with error code %s (%s)' % ('DecryptMessage', result, SEC_E(result)))
		
	_DecryptMessage = windll.Secur32.DecryptMessage
	_DecryptMessage.argtypes = [PCtxtHandle, PSecBufferDesc, ULONG, PULONG]
	_DecryptMessage.restype  = DWORD
	_DecryptMessage.errcheck  = errc
	
	secbuffers = []
	secbuffers.append(SecBuffer(token=token, buffer_type = SECBUFFER_TYPE.SECBUFFER_TOKEN))
	secbuffers.append(SecBuffer(token=data[:-1], buffer_type = SECBUFFER_TYPE.SECBUFFER_DATA))
	secbuffers.append(SecBuffer(token=data[-1:],buffer_type = SECBUFFER_TYPE.SECBUFFER_PADDING))
	
	
	data = SecBufferDesc(secbuffers)
	
	flags = ULONG()
	message_no = ULONG(message_no)

	res = _DecryptMessage(byref(ctx), byref(data), message_no, byref(flags))
	
	return data.Buffers

def GetSequenceNumberFromEncryptdataKerberos(ctx, unwrap=True):
	def errc(result, func, arguments):
		if SEC_E(result) == SEC_E.OK:
			return SEC_E(result)
		raise Exception('%s failed with error code %s (%s)' % ('EncryptMessage', result, SEC_E(result)))
		
	_EncryptMessage = windll.Secur32.EncryptMessage
	_EncryptMessage.argtypes = [PCtxtHandle, ULONG, PSecBufferDesc, ULONG]
	_EncryptMessage.restype  = DWORD
	_EncryptMessage.errcheck  = errc
	
	data = b'HAHAHAHAHAHAHAHA'

	secbuffers = []
	secbuffers.append(SecBuffer(token=b'\x00'*1024, buffer_type = SECBUFFER_TYPE.SECBUFFER_TOKEN)) #tthe correct size should be checked but that's another api call..
	secbuffers.append(SecBuffer(token=data, buffer_type = SECBUFFER_TYPE.SECBUFFER_DATA))
	
	data = SecBufferDesc(secbuffers)
	
	flags = ULONG(1)
	message_no = ULONG(0)

	res = _EncryptMessage(ctx, flags, byref(data), message_no)
	if unwrap is True:
		tok = GSSWrapToken.from_bytes(data.Buffers[0][1])
		return tok.SND_SEQ
	return data.Buffers[0][1]
	
def EncryptMessage(ctx, data, message_no = 0, fQOP = None):
	#raise NotImplementedError()
	def errc(result, func, arguments):
		if SEC_E(result) == SEC_E.OK:
			return SEC_E(result)
		raise Exception('%s failed with error code %s (%s)' % ('EncryptMessage', result, SEC_E(result)))
		
	_EncryptMessage = windll.Secur32.EncryptMessage
	_EncryptMessage.argtypes = [PCtxtHandle, ULONG, PSecBufferDesc, ULONG]
	_EncryptMessage.restype  = DWORD
	_EncryptMessage.errcheck  = errc
	
	secbuffers = []
	#secbuffers.append(SecBuffer(token=b'\x00'*1024, buffer_type = SECBUFFER_TYPE.SECBUFFER_STREAM_HEADER))
	secbuffers.append(SecBuffer(token=b'\x00'*1024, buffer_type = SECBUFFER_TYPE.SECBUFFER_TOKEN))
	secbuffers.append(SecBuffer(token=data, buffer_type = SECBUFFER_TYPE.SECBUFFER_DATA))
	secbuffers.append(SecBuffer(token =b'\x00'*1024,buffer_type = SECBUFFER_TYPE.SECBUFFER_PADDING))
	#secbuffers.append(SecBuffer(token = b'',buffer_type = SECBUFFER_TYPE.SECBUFFER_EMPTY))
	
	data = SecBufferDesc(secbuffers)
	
	flags = ULONG(1)
	message_no = ULONG(message_no)

	res = _EncryptMessage(ctx, flags, byref(data), message_no)
	return res, data.Buffers
	

	
# https://docs.microsoft.com/en-us/windows/desktop/api/sspi/nf-sspi-querycontextattributesa
def QueryContextAttributes(ctx, attr, sec_struct):
	#attr = SECPKG_ATTR enum
	def errc(result, func, arguments):
		if SEC_E(result) == SEC_E.OK:
			return SEC_E(result)
		raise Exception('%s failed with error code %s (%s)' % ('QueryContextAttributes', result, SEC_E(result)))
		
	_QueryContextAttributes = windll.Secur32.QueryContextAttributesW
	_QueryContextAttributes.argtypes = [PCtxtHandle, ULONG, PVOID]
	_QueryContextAttributes.restype  = DWORD
	_QueryContextAttributes.errcheck  = errc
	
	res = _QueryContextAttributes(byref(ctx), attr.value, byref(sec_struct))
	
	return


# https://docs.microsoft.com/en-us/windows/win32/api/sspi/nf-sspi-setcontextattributesw
def SetContextAttributes(ctx, attr, data):
	#attr = SECPKG_ATTR enum
	def errc(result, func, arguments):
		if SEC_E(result) == SEC_E.OK:
			return SEC_E(result)
		raise Exception('%s failed with error code %s (%s)' % ('SetContextAttributes', result, SEC_E(result)))
		
	_SetContextAttributes = windll.Secur32.SetContextAttributesW
	_SetContextAttributes.argtypes = [PCtxtHandle, ULONG, PVOID, ULONG]
	_SetContextAttributes.restype  = DWORD
	_SetContextAttributes.errcheck  = errc

	print('set data: %s' % data)
	print('set attr: %s' % attr)
	data_len = ULONG(len(data))
	data_buff = ctypes.create_string_buffer(data, len(data))
	
	res = _SetContextAttributes(byref(ctx), attr.value, data_buff, data_len)
	
	return





from asyauth.protocols.credssp.messages.asn1_structs import *

## this is not implemented yet!

"""
typedef struct _NTLM_REMOTE_SUPPLEMENTAL_CREDENTIAL {
      ULONG Version; 
      ULONG Flags;
      MSV1_0_CREDENTIAL_KEY_TYPE reserved;
      MSV1_0_CREDENTIAL_KEY reserved;
      ULONG reservedsize;
      [size_is(reservedSize)] UCHAR* reserved;
    } NTLM_REMOTE_SUPPLEMENTAL_CREDENTIAL; 




"""
import enum

class MSV1_0_CREDENTIAL_KEY_TYPE(enum.Enum):
	InvalidCredKey = 0 #        // reserved 
	IUMCredKey = 1 #             // reserved 
	DomainUserCredKey = 2 #     
	LocalUserCredKey = 3 #      // For internal use only - should never be present in MSV1_0_REMOTE_ENCRYPTED_SECRETS
	ExternallySuppliedCredKey = 4 # // reserved

class NRSC_FLAG(enum.IntFlag):
	LMOWF = 1
	NTOWF = 2
	CREDKEY_PRESENT = 8

class NTLM_REMOTE_SUPPLEMENTAL_CREDENTIAL:
	def __init__(self):
		self.Version:int = None
		self.Flags:int = None
		self.KeyType:MSV1_0_CREDENTIAL_KEY_TYPE = None
		self.Key:bytes = None
		self.reservedSize: int = None
		self.reserved6: bytes = None #size is 'reservedSize'

		
	def to_bytes(self):
		t  = self.Version.to_bytes(4, byteorder='little', signed = False)
		t += self.Flags.to_bytes(4, byteorder='little', signed = False)
		t += self.KeyType.value.to_bytes(4, byteorder='little', signed = False)
		t += self.Key
		t += len(self.reserved6).to_bytes(4, byteorder='little', signed = False)
		t += self.reserved6

		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return NTLM_REMOTE_SUPPLEMENTAL_CREDENTIAL.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = NTLM_REMOTE_SUPPLEMENTAL_CREDENTIAL()
		msg.Version = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.Flags = NRSC_FLAG(int.from_bytes(buff.read(4), byteorder='little', signed = False))
		msg.Key = buff.read(20)
		msg.KeyType = MSV1_0_CREDENTIAL_KEY_TYPE(int.from_bytes(buff.read(4), byteorder='little', signed = False))
		msg.reservedSize = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.reserved6 = buff.read(msg.reservedSize)
		return msg

	def __repr__(self):
		t = '==== NTLM_REMOTE_SUPPLEMENTAL_CREDENTIAL ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t



# TSRemoteGuardPackageCred
data = bytes.fromhex('3081dea00a04084e0054004c004d00a181cf0481cc0200ffff08000000d389fe0bc98d23bfef683a874e048c59000000000200000054000000ca7a124b7c282f0c714e025b3f5486310100000000000000000000000000000047e7674810c28f0cdb956d0aa1f4cac4005fb744b102871e16b207d789b3da815b9fac95ba79da02c2ba0e134472979f4b592621000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000')

x = TSRemoteGuardPackageCred.load(data)
print(x.native)

print(x.native['credBuffer'])

credbuff = NTLM_REMOTE_SUPPLEMENTAL_CREDENTIAL.from_bytes(x.native['credBuffer'])
print(credbuff)

print(credbuff.reserved6[:16])
print(credbuff.reserved6[16:32])
print(credbuff.reserved6[32:])

data = bytes.fromhex('30820157a003020104a382014e0482014a050407ff0000001c00000000681ec3f55cbce0b2ef037dbe8afa8f7a6dbc5e7439ff675453fbcdb7cfd0c5b4af942453acf92677f21be4dc829c9e553280c5f37c08c5db5453330328d2fb8e65dd3f1a4f934d6888645365fee5e65888219f6dcaf5020ecea4a91595c1c95b1e3229d1b389c2ee23876e1c89a2fa3b06610201fce4a2427c8a95d52234c2343fbaaeeedecd636a0dc8b3fad75b3be9736ab488183e0bdc793dcce61fc019811970c1d56a02fcd7827ea64f60e163adf01b47ae1641d48ac719ba77c78ec5818d6a09c41c38da69b359a63a602319ab735243acf6c18d40307290d622bf6546a4650494b481228fcaaa62393e29fd10e6d31646eed50fa658eb0d8454ccf88dc4bf859cd98feed69d17fafce30bd03a6e0ccdccd55ba7cc146b60648308cabd2ee989e711675e4150da79ea9070df22223d2df1e41ed687a6575a2ecbce')
x = TSRequest.load(data)
print(x.native)


from typing import List
from asyauth.common.credentials import UniCredential
from asyauth.common.constants import asyauthSecret, asyauthProtocol, asyauthSubProtocol
from asyauth.common.subprotocols import SubProtocol
from asyauth.common.subprotocols import SubProtocolNative
from asysocks.unicomm.common.target import UniTarget
from minikerberos.common.creds import KerberosCredential as KCRED
from minikerberos.pkinit import PKINIT
from asyauth.utils.paramprocessor import str_one, int_one, bool_one, int_list


class KerberosCredential(UniCredential):
	def __init__(self, secret, username, domain, stype:asyauthSecret, target:UniTarget = None, altname:str = None, altdomain:str = None, etypes:List[int] = [23,17,18], subprotocol:SubProtocol = SubProtocolNative(), certdata:str = None, keydata:str=None):
		UniCredential.__init__(
			self, 
			secret = secret,
			username = username,
			domain = domain,
			stype = stype,			
			protocol = asyauthProtocol.KERBEROS,
			subprotocol = subprotocol)
		
		self.etypes = etypes
		self.altname = altname
		self.altdomain = altdomain
		self.target = target
		self.certdata = certdata #can be file path or cert data
		self.keydata = keydata #can be file path or cert data

		self.dh_params = {
				'p' : int('00ffffffffffffffffc90fdaa22168c234c4c6628b80dc1cd129024e088a67cc74020bbea63b139b22514a08798e3404ddef9519b3cd3a431b302b0a6df25f14374fe1356d6d51c245e485b576625e7ec6f44c42e9a637ed6b0bff5cb6f406b7edee386bfb5a899fa5ae9f24117c4b1fe649286651ece65381ffffffffffffffff', 16),
				'g' : 2
			}

		if self.stype is asyauthSecret.NT:
			self.stype = asyauthSecret.RC4

		self.sanity_check()
	
	def sanity_check(self):
		if self.domain is None or self.domain == '':
			raise Exception('Kerberos credential must have domain set! %s' % str(self))
		if self.target is None or self.target.dc_ip is None:
			raise Exception('Kerberos credential target must have dc_ip set!')		

	def get_pkinit(self) -> PKINIT:
		if self.stype == asyauthSecret.CERTSTORE:
			return PKINIT.from_windows_certstore(self.settings.pfx12_file, certstore_name = 'MY', cert_serial = None, dh_params = self.dh_params, is_azure = self.is_azure)
		else:
			return PKINIT.from_pfx(self.settings.pfx12_file, self.settings.pfx12_file_pass, dh_params = self.dh_params, is_azure = self.is_azure)

	@staticmethod
	def get_url_params():
		return {
			'altname' : str_one,
			'altdomain' : str_one,
			'etype' : int_list,
			'dc' : str_one,
			'dns' : str_one,
			'certdata':str_one,
			'keydata':str_one,
		}

	def get_basetype_and_encoding(self):
		basetype = self.stype
		encoding = 'file'
		if self.stype.name.endswith('B64'):
			basetype = asyauthSecret(self.stype.name[:-3])
			encoding = 'b64'
		elif self.stype.name.endswith('HEX'):
			basetype = asyauthSecret(self.stype.name[:-3])
			encoding = 'hex'
		return basetype, encoding

	def to_ccred(self) -> KCRED:
		basetype, encoding = self.get_basetype_and_encoding()
		if basetype == asyauthSecret.KEYTAB:
			return KCRED.from_keytab(self.secret, self.username, self.domain, encoding=encoding)
		if basetype == asyauthSecret.KIRBI:
			return KCRED.from_kirbi(self.secret, encoding=encoding)
		if basetype == asyauthSecret.CCACHE:
			return KCRED.from_ccache(self.secret, self.username, self.domain, encoding=encoding)
		if basetype == asyauthSecret.PFX:
			return KCRED.from_pfx(self.keydata, self.secret, self.dh_params, self.altname, self.altdomain, encoding=encoding)
		if basetype == asyauthSecret.PFXSTR:
			return KCRED.from_pfx_string(self.keydata, self.secret, self.dh_params, self.altname, self.altdomain)

		res = KCRED()
		res.username = self.username
		res.domain = self.domain

		if self.stype in [asyauthSecret.PASSWORD, asyauthSecret.PW, asyauthSecret.PASS]:
			res.password = self.secret
		elif self.stype in [asyauthSecret.NT, asyauthSecret.RC4]:
			if isinstance(self.secret, str) is True:
				if len(self.secret) != 32:
					raise Exception('Incorrect RC4/NT key! %s' % self.secret)
			elif isinstance(self.secret, bytes):
				if len(self.secret) != 16:
					raise Exception('Incorrect RC4/NT key! %s' % self.secret)
			res.nt_hash = self.secret
			res.kerberos_key_rc4 = self.secret
		elif self.stype == asyauthSecret.AES:
			if isinstance(self.secret, str) is True:
				if len(self.secret) == 32:
					res.kerberos_key_aes_128 = self.secret
				elif len(self.secret) == 64:
					res.kerberos_key_aes_256 = self.secret
				else:
					raise Exception('Incorrect AES key! %s' % self.secret)
			elif isinstance(self.secret, bytes) is True:
				if len(self.secret) == 16:
					res.kerberos_key_aes_128 = self.secret
				elif len(self.secret) == 32:
					res.kerberos_key_aes_256 = self.secret
				else:
					raise Exception('Incorrect AES key! %s' % self.secret)
			else:
				raise Exception('Incorrect AES key! %s' % self.secret)
		elif self.stype == asyauthSecret.DES:
			if isinstance(self.secret, str) is True:
				if len(self.secret) != 16:
					raise Exception('Incorrect DES key! %s' % self.secret)
			elif isinstance(self.secret, bytearray) is True:
				if len(self.secret) != 8:
					raise Exception('Incorrect DES key! %s' % self.secret)
			res.kerberos_key_des = self.secret
		elif self.stype in [asyauthSecret.DES3, asyauthSecret.TDES]:
			if isinstance(self.secret, str) is True:
				if len(self.secret) != 24:
					raise Exception('Incorrect DES3 key! %s' % self.secret)
			elif isinstance(self.secret, bytearray) is True:
				if len(self.secret) != 12:
					raise Exception('Incorrect DES3 key! %s' % self.secret)
			res.kerberos_key_des3 = self.secret
		elif self.stype == asyauthSecret.NONE:
			res.nopreauth = True
		else:
			raise Exception('Missing/unknown stype! %s' % self.stype)

		return res

	def build_context(self):
		if self.subprotocol.type == asyauthSubProtocol.NATIVE:
			from asyauth.protocols.kerberos.client.native import KerberosClientNative
			return KerberosClientNative(self)

		elif self.subprotocol.type == asyauthSubProtocol.SSPI:
			from asyauth.protocols.kerberos.client.sspi import KerberosClientSSPI
			return KerberosClientSSPI(self)

		elif self.subprotocol.type == asyauthSubProtocol.SSPIPROXY:
			from asyauth.protocols.kerberos.client.sspiproxy import KerberosClientSSPIProxy
			return KerberosClientSSPIProxy(self)

		elif self.subprotocol.type == asyauthSubProtocol.WSNET:
			from asyauth.protocols.kerberos.client.wsnet import KerberosClientWSNET
			return KerberosClientWSNET(self)
		elif self.subprotocol.type == asyauthSubProtocol.CUSTOM:
			return self.subprotocol.factoryobj.build_context(self)
		else:
			raise Exception('Unsupported subprotocol "%s"' % self.subprotocol)





from asyauth import logger
from asyauth.common.winapi.constants import ISC_REQ
from asyauth.protocols.kerberos.gssapi import get_gssapi
from asyauth.protocols.kerberos.gssapismb import get_gssapi as gssapi_smb
from minikerberos.protocol.asn1_structs import AP_REP, EncAPRepPart, Ticket
from minikerberos.protocol.structures import ChecksumFlags
from minikerberos.gssapi.gssapi import GSSAPIFlags
from minikerberos.protocol.encryption import Key, _enctype_table
from minikerberos.aioclient import AIOKerberosClient
from asyauth.common.credentials.kerberos import KerberosCredential
from minikerberos.common.spn import KerberosSPN
import datetime
from minikerberos.protocol.constants import MESSAGE_TYPE
from minikerberos.protocol.asn1_structs import AP_REP, EncAPRepPart, EncryptedData, Ticket


class KerberosClientNative:
	def __init__(self, credential:KerberosCredential):
		self.credential = credential
		self.ccred = self.credential.to_ccred()
		
		self.kc = None
		self.session_key = None
		self.gssapi = None
		self.iterations = 0
		self.seq_number = 0
		self.from_ccache = False
	
		self.flags = \
			GSSAPIFlags.GSS_C_CONF_FLAG |\
			GSSAPIFlags.GSS_C_INTEG_FLAG |\
			GSSAPIFlags.GSS_C_REPLAY_FLAG |\
			GSSAPIFlags.GSS_C_SEQUENCE_FLAG
		
	
	def get_seq_number(self):
		"""
		Returns the initial sequence number. It is 0 by default, but can be adjusted during authentication, 
		by passing the 'seq_number' parameter in the 'authenticate' function
		"""
		return self.seq_number
	
	def signing_needed(self):
		"""
		Checks if integrity protection was negotiated
		"""
		return GSSAPIFlags.GSS_C_INTEG_FLAG in self.flags
	
	def encryption_needed(self):
		"""
		Checks if confidentiality flag was negotiated
		"""
		return GSSAPIFlags.GSS_C_CONF_FLAG in self.flags
				
	async def sign(self, data, message_no, direction = 'init'):
		"""
		Signs a message. 
		"""
		return self.gssapi.GSS_GetMIC(data, message_no, direction = direction)	
		
	async def encrypt(self, data, message_no, *args, **kwargs):
		"""
		Encrypts a message. 
		"""
		data, eeee  = self.gssapi.GSS_Wrap(data, message_no, *args, **kwargs)
		return data, eeee 
		
	async def decrypt(self, data, message_no, *args, **kwargs):
		"""
		Decrypts message. Also performs integrity checking.
		"""

		return self.gssapi.GSS_Unwrap(data, message_no, *args, **kwargs)
	
	def get_session_key(self):
		return self.session_key.contents

	def iscreq_to_gssapiflags(self, flags):
		if flags is None:
			return self.flags
		kflags = GSSAPIFlags.GSS_C_CONF_FLAG |\
			GSSAPIFlags.GSS_C_INTEG_FLAG |\
			GSSAPIFlags.GSS_C_REPLAY_FLAG |\
			GSSAPIFlags.GSS_C_SEQUENCE_FLAG
		if ISC_REQ.INTEGRITY in flags:
			kflags |= GSSAPIFlags.GSS_C_INTEG_FLAG
		else:
			kflags &= ~GSSAPIFlags.GSS_C_INTEG_FLAG
		if ISC_REQ.CONFIDENTIALITY in flags:
			kflags |= GSSAPIFlags.GSS_C_CONF_FLAG
		else:
			kflags &= ~GSSAPIFlags.GSS_C_CONF_FLAG
		if ISC_REQ.REPLAY_DETECT in flags:
			kflags |= GSSAPIFlags.GSS_C_REPLAY_FLAG
		else:
			kflags &= ~GSSAPIFlags.GSS_C_REPLAY_FLAG
		if ISC_REQ.SEQUENCE_DETECT in flags:
			kflags |= GSSAPIFlags.GSS_C_SEQUENCE_FLAG
		else:
			kflags &= ~GSSAPIFlags.GSS_C_SEQUENCE_FLAG
		if ISC_REQ.USE_DCE_STYLE in flags:
			kflags |= GSSAPIFlags.GSS_C_DCE_STYLE
		else:
			kflags &= ~GSSAPIFlags.GSS_C_DCE_STYLE
		if ISC_REQ.MUTUAL_AUTH in flags:
			kflags |= GSSAPIFlags.GSS_C_MUTUAL_FLAG
		else:
			kflags &= ~GSSAPIFlags.GSS_C_MUTUAL_FLAG
		return kflags
		
	
	async def authenticate(self, authData, flags = None, seq_number = 0, cb_data = None, spn = None):
		"""
		This function is called (multiple times depending on the flags) to perform authentication. 
		"""
		try:
			self.flags = self.iscreq_to_gssapiflags(flags)

			if spn is None:
				raise Exception("SPN is needed for kerberos!")
			else:
				spn = KerberosSPN.from_target_string(spn)

			if self.kc is None:
				self.kc = AIOKerberosClient(self.ccred, self.credential.target)

			if self.iterations == 0:
				self.seq_number = 0
				self.iterations += 1
				
				try:
					#check TGS first, maybe ccache already has what we need
					for target in self.ccred.ccache.list_targets():
						# just printing this to debug...
						logger.debug('CCACHE SPN record: %s' % target)
					tgs, encpart, self.session_key = await self.kc.get_TGS(spn)
					
					self.from_ccache = True
				except:
					tgt = await self.kc.get_TGT(override_etype = self.credential.etypes)
					tgs, encpart, self.session_key = await self.kc.get_TGS(spn)#, override_etype = self.preferred_etypes)
				
				ap_opts = []
				if GSSAPIFlags.GSS_C_MUTUAL_FLAG in self.flags or GSSAPIFlags.GSS_C_DCE_STYLE in self.flags:
					if GSSAPIFlags.GSS_C_MUTUAL_FLAG in self.flags:
						ap_opts.append('mutual-required')
					if self.from_ccache is False:
						apreq = self.kc.construct_apreq(
							tgs, 
							encpart, 
							self.session_key, 
							flags = self.flags, 
							seq_number = self.seq_number, 
							ap_opts=ap_opts, 
							cb_data = cb_data
						)
					else:
						apreq = self.kc.construct_apreq_from_ticket(
							Ticket(tgs['ticket']).dump(), 
							self.session_key, 
							tgs['crealm'], tgs['cname']['name-string'][0], 
							flags = self.flags, 
							seq_number = self.seq_number, 
							ap_opts = ap_opts, 
							cb_data = cb_data
						)
					return apreq, True, None
				
				else:
					#not mutual nor dce auth will take one step only
					if self.from_ccache is False:
						apreq = self.kc.construct_apreq(
							tgs, 
							encpart, 
							self.session_key, 
							flags = self.flags, 
							seq_number = self.seq_number, 
							ap_opts=ap_opts, 
							cb_data = cb_data)
					else:
						apreq = self.kc.construct_apreq_from_ticket(
							Ticket(tgs['ticket']).dump(), 
							self.session_key, 
							tgs['crealm'], 
							tgs['cname']['name-string'][0], 
							flags = self.flags, 
							seq_number = self.seq_number, 
							ap_opts = ap_opts, 
							cb_data = cb_data
						)
					
					
					self.gssapi = get_gssapi(self.session_key)
					return apreq, False, None

			else:
				self.seq_number = seq_number

				aprep = AP_REP.load(authData).native
				cipher = _enctype_table[int(aprep['enc-part']['etype'])]()
				cipher_text = aprep['enc-part']['cipher']
				temp = cipher.decrypt(self.session_key, 12, cipher_text)
					
				enc_part = EncAPRepPart.load(temp).native
				cipher = _enctype_table[int(enc_part['subkey']['keytype'])]()
					
				now = datetime.datetime.now(datetime.timezone.utc)
				apreppart_data = {}
				apreppart_data['cusec'] = now.microsecond
				apreppart_data['ctime'] = now.replace(microsecond=0)
				apreppart_data['seq-number'] = enc_part['seq-number']
				#print('seq %s' % enc_part['seq-number'])
				#self.seq_number = 0 #enc_part['seq-number']
					
				apreppart_data_enc = cipher.encrypt(self.session_key, 12, EncAPRepPart(apreppart_data).dump(), None)
					
				#overriding current session key
				self.session_key = Key(cipher.enctype, enc_part['subkey']['keyvalue'])
					
				ap_rep = {}
				ap_rep['pvno'] = 5 
				ap_rep['msg-type'] = MESSAGE_TYPE.KRB_AP_REP.value
				ap_rep['enc-part'] = EncryptedData({'etype': self.session_key.enctype, 'cipher': apreppart_data_enc}) 
					
				token = AP_REP(ap_rep).dump()
				if GSSAPIFlags.GSS_C_DCE_STYLE in self.flags:
					self.gssapi = gssapi_smb(self.session_key)
				else:
					self.gssapi = get_gssapi(self.session_key)
				self.iterations += 1
				return token, True, None
		
		except Exception as e:
			return None, None, e



import os
from asyauth import logger
from asyauth.protocols.spnego.client.native import SPNEGOClientNative
from asyauth.protocols.credssp.messages.asn1_structs import NegoDatas, \
	NegoData, TSRequest, TSRequest, TSPasswordCreds, TSCredentials, TSRemoteGuardCreds, \
	TSRemoteGuardPackageCred
from hashlib import sha256
from asn1crypto.x509 import Certificate
from asyauth.common.constants import asyauthSecret

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-cssp/385a7489-d46b-464c-b224-f7340e308a5c

class CredSSPClientNative:
	def __init__(self, credential):
		self.credential = credential
		self.auth_ctx:SPNEGOClientNative = None
		self.cred = None
		self.version = 4 #ver 5 and 6 not working TODO
		self.nonce = os.urandom(32)
		self.__internal_auth_continue = True
		self.seqno = 0

	@staticmethod
	def certificate_to_pubkey(certdata):
		# credSSP auth requires knowledge of the server's public key
		cert = Certificate.load(certdata)
		return cert['tbs_certificate']['subject_public_key_info']['public_key'].dump()[5:] #why?

	def get_extra_info(self):
		return self.auth_ctx.get_extra_info()

	def get_active_credential(self):
		return self.auth_ctx.get_active_credential()

	async def authenticate(self, token, flags = None, certificate = None, spn = None, remote_credguard = False):
		try:
			# currently only SSPI supported

			if certificate is not None:
				pubkey = CredSSPClientNative.certificate_to_pubkey(certificate)
			
			if token is None:
				# initial auth
				returndata, self.__internal_auth_continue, err = await self.auth_ctx.authenticate(token, flags = flags, spn = spn)
				if err is not None:
					raise err
					
				negotoken = {
					'negoToken' : returndata
				}
				retoken = {
					'version' : self.version,
					'negoTokens' : NegoDatas([NegoData(negotoken)])
				}
				return TSRequest(retoken).dump(), True, None
			else:
				if self.__internal_auth_continue is True:
					tdata = TSRequest.load(token)
					if tdata.native['version'] < self.version:
						logger.debug('[CREDSSP] Server supports version %s which is smaller than our supported version %s' % (tdata.native['version'], self.version))
						self.version = tdata.native['version']
					if tdata.native['negoTokens'] is None:
						raise Exception('SSPI auth not supported by server')
					sspitoken = tdata.native['negoTokens'][0]['negoToken']
					returndata, self.__internal_auth_continue, err = await self.auth_ctx.authenticate(sspitoken, flags = flags, spn = spn)
					if err is not None:
						raise err
						
					negotoken = {
						'negoToken' : returndata
					}
					retoken = {
						'version' : self.version,
					}
					if returndata is not None:
						retoken['negoTokens'] = NegoDatas([NegoData(negotoken)])
						
						
					if self.__internal_auth_continue is False:
						self.seqno = self.auth_ctx.get_internal_seq() #spnego might increate a seq number when signing
						if self.version in [5,6]:
							ClientServerHashMagic = b"CredSSP Client-To-Server Binding Hash\x00"
							ClientServerHash = sha256(ClientServerHashMagic + self.nonce + pubkey).digest()
							sealedMessage, signature = await self.auth_ctx.encrypt(ClientServerHash, self.seqno)
							self.seqno += 1
							retoken['pubKeyAuth'] = signature+sealedMessage
							retoken['clientNonce'] = self.nonce
							
						elif self.version in [2,3,4]:
							sealedMessage, signature = await self.auth_ctx.encrypt(pubkey, self.seqno)
							self.seqno += 1
							retoken['pubKeyAuth'] = signature+sealedMessage
						
					return TSRequest(retoken).dump(), True, None
				else:
					# sub-level auth protocol finished, now for the other stuff
						
					# waiting for server to reply with the re-encrypted verification string + b'\x01'
					tdata = TSRequest.load(token).native
					if tdata['errorCode'] is not None:
						raise Exception('CredSSP - Server sent an error! Code: %s' % hex(tdata['errorCode'] & (2**32-1)))
					if tdata['pubKeyAuth'] is None:
						raise Exception('Missing pubKeyAuth')
					verification_data, _ = await self.auth_ctx.decrypt(tdata['pubKeyAuth'], 0)
					#print('DEC: %s' % verification_data)

					# at this point the verification should be implemented
					# TODO: maybe later...

					
					# sending credentials
					if remote_credguard is False:
						creds = {
							'domainName' : b'',
							'userName'   : b'',
							'password'   : b'',
						}
						cred = self.auth_ctx.get_active_credential()
						if cred.stype == asyauthSecret.PASSWORD:
							creds = {
								'domainName' : cred.domain.encode('utf-16-le'),
								'userName'   : cred.username.encode('utf-16-le'),
								'password'   : cred.secret.encode('utf-16-le'),
							}
								 
										
						res = TSPasswordCreds(creds)
						res = TSCredentials({'credType': 1, 'credentials': res.dump()})
						sealedMessage, signature = await self.auth_ctx.encrypt(res.dump(), self.seqno) #seq number must be incremented here..
						self.seqno += 1
						retoken = {
							'version' : self.version,
							'authInfo' : signature+sealedMessage
						}
						return TSRequest(retoken).dump(), False, None
					else:
						credBuffer = b''
						data = {
							'packageName' : 'KERBEROS'.encode('utf-16-le'), #dont forget the encoding!
							'credBuffer' : credBuffer
						}

						remcredguardcreds = TSRemoteGuardCreds({
							'logonCred' : TSRemoteGuardPackageCred(data),
							#'supplementalCreds' : [TSRemoteGuardPackageCred(xxxx)]
						})

						#print(remcredguardcreds)
						sealedMessage, signature = await self.auth_ctx.encrypt(remcredguardcreds.dump(), self.seqno) #seq number must be incremented here..
						self.seqno += 1
						retoken = {
							'version' : self.version,
							'authInfo' : signature+sealedMessage
						}
						return TSRequest(retoken).dump(), False, None						
		except Exception as e:
			return None, None, e


import enum

class asyauthProtocol(enum.Enum):
	NONE = 'NONE'
	SIMPLE = 'SIMPLE'
	PLAIN = 'PLAIN'
	SICILY = 'SICILY'
	NTLM = 'NTLM'
	KERBEROS = 'KERBEROS'
	CREDSSP = 'CREDSSP'
	SPNEGO = 'SPNEGO'
	SPNEGOEX = 'SPNEGOEX'
	SSL = 'SSL' # SSL/TLS schannel later

class asyauthSubProtocol(enum.Enum):
	NATIVE = 'NATIVE'
	SSPI = 'SSPI'
	WSNET = 'WSNET'
	SSPIPROXY = 'SSPIPROXY'
	CUSTOM = 'CUSTOM'

class asyauthSecret(enum.Enum):
	NONE = 'NONE'
	PASSWORD = 'PASSWORD'
	PASS = 'PASS'
	PW = 'PW'
	PWB64 = 'PWB64'
	PWPROMPT = 'PWPROMPT'
	PWHEX = 'PWHEX'
	LM = 'LM'
	NT = 'NT'
	RC4 = 'RC4'
	AES = 'AES'
	CCACHE = 'CCACHE'
	CCACHEHEX = 'CCACHEHEX'
	CCACHEB64 = 'CCACHEB64'
	KEYTAB = 'KEYTAB'
	KEYTABHEX = 'KEYTABHEX'
	KEYTABB64 = 'KEYTABB64'
	PFX = 'PFX'
	PFXB64 = 'PFXB64'
	PFXHEX = 'PFXHEX'
	PFXSTR = 'PFXSTR'
	PEM = 'PEM'
	CERTSTORE = 'CERTSTORE'
	KIRBI = 'KIRBI'
	KIRBIHEX = 'KIRBIHEX'
	KIRBIB64 = 'KIRBIB64'
	DES = 'DES'
	TDES = 'TDES'
	DES3 = 'DES3'
	SSHPRIVKEY = 'SSHPRIVKEY'
	SSHPRIVKEYSTR = 'SSHPRIVKEYSTR'
	SSHPRIVKEYB64 = 'SSHPRIVKEYB64'


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

import logging

logger = logging.getLogger(__name__)
handler = logging.StreamHandler()
formatter = logging.Formatter(
        '%(asctime)s %(name)-12s %(levelname)-8s %(message)s')
handler.setFormatter(formatter)
logger.addHandler(handler)
logger.setLevel(logging.INFO)
logger.propagate = False



from re import sub
from asyauth.common.constants import asyauthSubProtocol
from urllib.parse import urlparse, parse_qs

class SubProtocol:
	def __init__(self, type:asyauthSubProtocol):
		self.type = type
	
	@staticmethod
	def from_url_params(subprotocol:asyauthSubProtocol, query_params):
		if subprotocol == asyauthSubProtocol.NATIVE:
			return SubProtocolNative.from_url_params(query_params)
		if subprotocol == asyauthSubProtocol.SSPI:
			return SubProtocolSSPI.from_url_params(query_params)
		if subprotocol == asyauthSubProtocol.SSPIPROXY:
			return SubProtocolSSPIProxy.from_url_params(query_params)
		if subprotocol == asyauthSubProtocol.WSNET:
			return SubProtocolWSNet.from_url_params(query_params)
		raise Exception("asyauth unknown subprotocol %s" % subprotocol)

from asyauth.common.subprotocols.native import SubProtocolNative
from asyauth.common.subprotocols.sspi import SubProtocolSSPI
from asyauth.common.subprotocols.sspiproxy import SubProtocolSSPIProxy
from asyauth.common.subprotocols.wsnet import SubProtocolWSNet


__all__ = ['SubProtocolNative', 'SubProtocolSSPI', 'SubProtocolSSPIProxy', 'SubProtocolWSNet']





from asn1crypto.core import Integer, ObjectIdentifier, Sequence, SequenceOf, Enumerated, GeneralString, OctetString, BitString, Choice, Any, Boolean

TAG = 'explicit'

# class
UNIVERSAL = 0
APPLICATION = 1
CONTEXT = 2

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-cssp/9664994d-0784-4659-b85b-83b8d54c2336
#  NegoData ::= SEQUENCE OF SEQUENCE {
#         negoToken [0] OCTET STRING
# }

class NegoData(Sequence):
	_fields = [
		('negoToken', OctetString, {'explicit': 0}),]

class NegoDatas(SequenceOf):
	_child_spec = NegoData


# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-cssp/6aac4dea-08ef-47a6-8747-22ea7f6d8685
#  TSRequest ::= SEQUENCE {
#         version    [0] INTEGER,
#         negoTokens [1] NegoData  OPTIONAL,
#         authInfo   [2] OCTET STRING OPTIONAL,
#         pubKeyAuth [3] OCTET STRING OPTIONAL,
#         errorCode  [4] INTEGER OPTIONAL,
#         clientNonce [5] OCTET STRING OPTIONAL
# }

class TSRequest(Sequence):
	_fields = [
		('version', Integer, {'explicit': 0}),
		('negoTokens', NegoDatas, {'explicit': 1, 'optional': True}),
		('authInfo', OctetString, {'explicit': 2, 'optional': True}),
		('pubKeyAuth', OctetString, {'explicit': 3, 'optional': True}),
		('errorCode', Integer, {'explicit': 4, 'optional': True}),
		('clientNonce', OctetString, {'explicit': 5, 'optional': True}),
]



# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-cssp/94a1ab00-5500-42fd-8d3d-7a84e6c2cf03
# TSCredentials ::= SEQUENCE {
#         credType    [0] INTEGER,
#         credentials [1] OCTET STRING
# }

class TSCredentials(Sequence):
	_fields = [
		('credType', Integer, {'explicit': 0}),
		('credentials', OctetString, {'explicit': 1}),
]


# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-cssp/17773cc4-21e9-4a75-a0dd-72706b174fe5
# TSPasswordCreds ::= SEQUENCE {
#         domainName  [0] OCTET STRING,
#         userName    [1] OCTET STRING,
#         password    [2] OCTET STRING
# }

class TSPasswordCreds(Sequence):
	_fields = [
		('domainName', OctetString, {'explicit': 0}),
		('userName', OctetString, {'explicit': 1}),
		('password', OctetString, {'explicit': 2}),
]

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-cssp/34ee27b3-5791-43bb-9201-076054b58123
# TSCspDataDetail ::= SEQUENCE {
#         keySpec       [0] INTEGER,
#         cardName      [1] OCTET STRING OPTIONAL,
#         readerName    [2] OCTET STRING OPTIONAL,
#         containerName [3] OCTET STRING OPTIONAL,
#         cspName       [4] OCTET STRING OPTIONAL
# }

class TSCspDataDetail(Sequence):
	_fields = [
		('keySpec', Integer, {'explicit': 0}),
		('cardName', OctetString, {'explicit': 1,'optional': True}),
		('readerName', OctetString, {'explicit': 2,'optional': True}),
		('containerName', OctetString, {'explicit': 3,'optional': True}),
		('cspName', OctetString, {'explicit': 4,'optional': True }),
]


# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-cssp/4251d165-cf01-4513-a5d8-39ee4a98b7a4
# TSSmartCardCreds ::= SEQUENCE {
#         pin         [0] OCTET STRING,
#         cspData     [1] TSCspDataDetail,
#         userHint    [2] OCTET STRING OPTIONAL,
#         domainHint  [3] OCTET STRING OPTIONAL
# }
#

class TSSmartCardCreds(Sequence):
	_fields = [
		('pin', OctetString, {'explicit': 0}),
		('cspData', TSCspDataDetail, {'explicit': 1}),
		('userHint', OctetString, {'explicit': 2, 'optional': True}),
		('domainHint', OctetString, {'explicit': 3, 'optional': True}),
]

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-cssp/173eee44-1a2c-463f-b909-c15db01e68d7
# TSRemoteGuardPackageCred ::= SEQUENCE{
#     packageName [0] OCTET STRING,
#     credBuffer  [1] OCTET STRING,
# }

class TSRemoteGuardPackageCred(Sequence):
	_fields = [
		('packageName', OctetString, {'explicit': 0}),
		('credBuffer', OctetString, {'explicit': 1}),
]


# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-cssp/7ef8229c-44ea-4c1b-867f-00369b882b38
# TSRemoteGuardCreds ::= SEQUENCE{
#     logonCred           [0] TSRemoteGuardPackageCred,
#     supplementalCreds   [1] SEQUENCE OF TSRemoteGuardPackageCred OPTIONAL,
# }


class TSRemoteGuardPackageCreds(SequenceOf):
	_child_spec = TSRemoteGuardPackageCred

class TSRemoteGuardCreds(Sequence):
	_fields = [
		('logonCred', TSRemoteGuardPackageCred, {'explicit': 0}),
		('supplementalCreds', TSRemoteGuardPackageCreds, {'explicit': 1, 'optional':True}),
]



# Kudos:
# Parts of this code was inspired by the following project by @rubin_mor
# https://github.com/morRubin/AzureADJoinedMachinePTC
# 

# TODO: code needs cleanup, it is still in beta
# TODO: add integrity checks and check certificate of the server
# TODO: code currently supports RSA+DH+SHA1 , add support for other mechanisms

import os
import base64

from asyauth.common.constants import asyauthSecret
from asyauth.common.credentials.kerberos import KerberosCredential
from asyauth.protocols.spnegoex.protocol.messages import MESSAGE_TYPE, PKU2U_TOKEN_TYPE, generate_verify, generate_initiator_metadata, generate_init_nego, generate_ap_req, negoexts_parse_bytes
from asyauth.protocols.kerberos.gssapismb import get_gssapi
from minikerberos.pkinit import PKINIT
from asysocks.unicomm.common.target import UniTarget


class SPNEGOEXClientNative:
	def __init__(self, credential:KerberosCredential):
		self.credential = credential
		self.target:UniTarget = None
		self.pkinit:PKINIT = self.credential.get_pkinit()
		self.gssapi = None
		self.is_azure = False

		self._convid = os.urandom(16)
		self._msgctr = 0
		self._krb_finished_data = b''
		self._msgs = b''
		self.session_key_data = None
		self.xxxxx = None

		self.iteractions = 0

	def get_session_key(self):
		return self.session_key.contents

	def __get_metadata(self):
		return generate_initiator_metadata(self._msgctr, self._convid, self.pkinit.get_metadata(target = self.credential.target.get_hostname_or_ip()))
	
	async def sign(self, data, message_no, direction = 'init'):
		return self.gssapi.GSS_GetMIC(data, message_no)	
		
	async def encrypt(self, data, message_no):
		return self.gssapi.GSS_Wrap(data, message_no)
		
	async def decrypt(self, data, message_no, direction='init', auth_data=None):
		return self.gssapi.GSS_Unwrap(data, message_no, direction=direction, auth_data=auth_data)
	
	async def authenticate(self, authData, flags = None, spn = None):
		if self.iteractions == 0:
			self.setup()
			self.iteractions += 1
			#authdata should be 0 at this point

			if self.is_azure is True:
				# kerberos service is on the same ip
				asreq = self.pkinit.build_asreq(target = self.credential.target.get_hostname_or_ip(), kdcopts = ['forwardable','renewable','proxiable', 'canonicalize'])
			else:
				if self.credential.target.dc_ip is None:
					raise Exception('DC IP must be set for kerberos auth!')
				asreq = self.pkinit.build_asreq(kdcopts = ['forwardable','renewable','proxiable', 'canonicalize'])

			negodata = generate_init_nego(self._msgctr, self._convid)
			self._msgctr += 1
			metadata = self.__get_metadata()
			self._msgctr += 1
			ap_req, token_raw = generate_ap_req(self._msgctr, self._convid, asreq, PKU2U_TOKEN_TYPE.KRB_AS_REQ)
			self._krb_finished_data += token_raw # for the checksum calc...
			self._msgctr += 1
			msg = negodata + metadata + ap_req
			self._msgs += msg

			return msg, True, None

		elif self.iteractions == 1:
			from minikerberos.protocol.encryption import Enctype, _checksum_table, _enctype_table, Key
			self.iteractions += 1
			
			self._msgs += authData
			msgs = negoexts_parse_bytes(authData)
			self._msgctr += len(msgs)
			#print(msgs[MESSAGE_TYPE.CHALLENGE].Exchange.inner_token.native)
			as_rep = msgs[MESSAGE_TYPE.CHALLENGE].Exchange.inner_token.native
			self._krb_finished_data += msgs[MESSAGE_TYPE.CHALLENGE].exchange_data_raw # for the checksum calc...
			encasrep, session_key, cipher = self.pkinit.decrypt_asrep(as_rep)

			self.xxxxx = session_key

			self.session_key_data = {}
			self.session_key_data['keytype'] = Enctype.AES256
			self.session_key_data['keyvalue'] = os.urandom(32)
			subkey_cipher = _enctype_table[self.session_key_data['keytype']]
			subkey_key = Key(subkey_cipher.enctype, self.session_key_data['keyvalue'])
			subkey_checksum = _checksum_table[16] # ChecksumTypes.hmac_sha1_96_aes256

			ap_req = self.pkinit.build_apreq(as_rep, session_key, cipher, self.session_key_data, self._krb_finished_data)

			ap_req_msg, _ = generate_ap_req(self._msgctr, self._convid, ap_req, PKU2U_TOKEN_TYPE.KRB_AP_REQ)
			#print(ap_req_msg.hex())
			self._msgctr += 1
			checksum_final = subkey_checksum.checksum(subkey_key, 25, self._msgs + ap_req_msg )
			verify_msg = generate_verify(self._msgctr, self._convid, checksum_final,  16)
			self._msgctr += 1

			ret_msg = ap_req_msg + verify_msg
			self._msgs += ret_msg

			return ret_msg, True, None


		elif self.iteractions == 2:
			from minikerberos.protocol.encryption import Enctype, _checksum_table, _enctype_table, Key
			from minikerberos.protocol.asn1_structs import EncAPRepPart

			#input('aaaaaaaaaaaaaa')
			self.iteractions += 1
			self._msgs += authData
			msgs = negoexts_parse_bytes(authData)
			self._msgctr += len(msgs)
			ap_rep = msgs[MESSAGE_TYPE.CHALLENGE].Exchange.inner_token.native
			#print(ap_rep)

			#self.xxxxx

			cipher = _enctype_table[int(ap_rep['enc-part']['etype'])]()
			cipher_text = ap_rep['enc-part']['cipher']
			subkey_key = Key(cipher.enctype, self.xxxxx.contents)
			temp = cipher.decrypt(subkey_key, 12, cipher_text)
			enc_part = EncAPRepPart.load(temp).native
			#print(enc_part)
			
			cipher = _enctype_table[int(enc_part['subkey']['keytype'])]()
			self.session_key = Key(cipher.enctype, enc_part['subkey']['keyvalue'])
			self.gssapi = get_gssapi(self.session_key)

			return None, False, None





import enum
from asyauth.common.winapi.constants import SECPKG_ATTR, SECPKG_CRED, ISC_REQ, SEC_E, SSPIResult
from asyauth.common.winapi.functiondefs import AcquireCredentialsHandle, InitializeSecurityContext,\
	DecryptMessage, EncryptMessage, SecPkgContext_SessionKey, QueryContextAttributes,\
	GetSequenceNumberFromEncryptdataKerberos
from minikerberos.protocol.asn1_structs import AP_REQ
from asyauth.common.winapi.token import InitialContextToken

class SSPIPackage(enum.Enum):
	NTLM = 'NTLM'
	KERBEROS = 'KERBEROS'
	DIGEST = 'DIGEST'
	NEGOTIATE = 'NEGOTIATE'
	CREDSSP = 'CREDSSP'
	SCHANNEL = 'SCHANNEL'

# first you must call initialize!
class WinSSPI:
	def __init__(self, package:SSPIPackage):
		self.cred_struct = None
		self.context = None
		self.package = package
		self.flags = None

	def get_seq_number(self):
		#only call it once!
		return GetSequenceNumberFromEncryptdataKerberos(self.context)
		
	def get_session_key(self):
		sec_struct = SecPkgContext_SessionKey()
		QueryContextAttributes(self.context, SECPKG_ATTR.SESSION_KEY, sec_struct)
		return sec_struct.Buffer
		
	def acquire_handle(self, client_name, target_name, flags = SECPKG_CRED.BOTH):
		self.cred_struct = AcquireCredentialsHandle(client_name, self.package.value, target_name, flags)
		
	def initialize_context(self, target, token_data = None, flags = ISC_REQ.INTEGRITY | ISC_REQ.CONFIDENTIALITY | ISC_REQ.SEQUENCE_DETECT | ISC_REQ.REPLAY_DETECT, cb_data = None):
		res, self.context, data, self.flags, expiry = InitializeSecurityContext(self.cred_struct, target, token = token_data, ctx = self.context, flags = flags, cb_data=cb_data)
		if res == SEC_E.OK:
			return SSPIResult.OK, data, self.flags, expiry
		else:
			return SSPIResult.CONTINUE, data, self.flags, expiry
			
	def _unwrap(self, data, message_no = 0):
		data_buff = DecryptMessage(self.context, data, message_no)
		return data_buff
		
	def _wrap(self, data, message_no = 0):
		data_buff = EncryptMessage(self.context, data, message_no)
		return data_buff

	def get_ticket_for_spn(self, target_name, flags = None, token_data = None, client_name = None, cb_data = None):
		try:
			self.target_name = target_name
			if not self.cred_struct:
				self.acquire_handle(client_name, self.target_name)
			
			if ISC_REQ.USE_DCE_STYLE in flags:
				res, self.context, data, outputflags, expiry = InitializeSecurityContext(self.cred_struct, self.target_name, token = token_data, ctx = self.context, flags = flags, cb_data=cb_data)
				return data, outputflags, None
			else:
				res, self.context, data, outputflags, expiry = InitializeSecurityContext(self.cred_struct, self.target_name, token = None, ctx = self.context, flags = flags, cb_data=cb_data)
				token = InitialContextToken.load(data) #this is disgusting :(
				return AP_REQ(token.native['innerContextToken']).dump(), outputflags, None #this is the AP_REQ
		except Exception as e:
			return None, None, e


import enum

# https://msdn.microsoft.com/en-us/library/cc236650.aspx
class NegotiateFlags(enum.IntFlag):
	NEGOTIATE_56   = 0x80000000
	NEGOTIATE_KEY_EXCH   = 0x40000000 # needed for encryption/singing key
	NEGOTIATE_128   = 0x20000000
	r1  = 0x10000000
	r2  = 0x8000000
	r3  = 0x4000000
	NEGOTIATE_VERSION   = 0x2000000
	r4  = 0x1000000
	NEGOTIATE_TARGET_INFO   = 0x800000
	REQUEST_NON_NT_SESSION_KEY   = 0x400000
	r5  = 0x200000
	NEGOTIATE_IDENTIFY   = 0x100000
	NEGOTIATE_EXTENDED_SESSIONSECURITY   = 0x80000
	r6  = 0x40000
	TARGET_TYPE_SERVER   = 0x20000
	TARGET_TYPE_DOMAIN   = 0x10000
	NEGOTIATE_ALWAYS_SIGN   = 0x8000
	r7  = 0x4000
	NEGOTIATE_OEM_WORKSTATION_SUPPLIED   = 0x2000
	NEGOTIATE_OEM_DOMAIN_SUPPLIED   = 0x1000
	J   = 0x800
	r8  = 0x400
	NEGOTIATE_NTLM   = 0x200
	r9  = 0x100
	NEGOTIATE_LM_KEY   = 0x80
	NEGOTIATE_DATAGRAM   = 0x40
	NEGOTIATE_SEAL   = 0x20
	NEGOTIATE_SIGN   = 0x10
	r10 = 0x8
	REQUEST_TARGET   = 0x4
	NTLM_NEGOTIATE_OEM   = 0x2
	NEGOTIATE_UNICODE   = 0x1
		
NegotiateFlagExp = {
	NegotiateFlags.NEGOTIATE_56   : 'requests 56-bit encryption. If the client sends NTLMSSP_NEGOTIATE_SEAL or NTLMSSP_NEGOTIATE_SIGN with NTLMSSP_NEGOTIATE_56 to the server in the NEGOTIATE_MESSAGE, the server MUST return NTLMSSP_NEGOTIATE_56 to the client in the CHALLENGE_MESSAGE.   Otherwise it is ignored. If both NTLMSSP_NEGOTIATE_56 and NTLMSSP_NEGOTIATE_128 are requested and supported by the client and server, NTLMSSP_NEGOTIATE_56 and NTLMSSP_NEGOTIATE_128 will both be returned to the client. Clients and servers that set NTLMSSP_NEGOTIATE_SEAL SHOULD set NTLMSSP_NEGOTIATE_56 if it is supported. An alternate name for this field is NTLMSSP_NEGOTIATE_56.',
	NegotiateFlags.NEGOTIATE_KEY_EXCH   : 'requests an explicit key exchange. This capability SHOULD be used because it improves security for message integrity or confidentiality. See sections 3.2.5.1.2, 3.2.5.2.1, and 3.2.5.2.2 for details. An alternate name for this field is NTLMSSP_NEGOTIATE_KEY_EXCH.',
	NegotiateFlags.NEGOTIATE_128  : 'requests 128-bit session key negotiation. An alternate name for this field is NTLMSSP_NEGOTIATE_128. If the client sends NTLMSSP_NEGOTIATE_128 to the server in the NEGOTIATE_MESSAGE, the server MUST return NTLMSSP_NEGOTIATE_128 to the client in the CHALLENGE_MESSAGE only if the client sets NTLMSSP_NEGOTIATE_SEAL or NTLMSSP_NEGOTIATE_SIGN. Otherwise it is ignored. If both NTLMSSP_NEGOTIATE_56 and NTLMSSP_NEGOTIATE_128 are requested and supported by the client and server, NTLMSSP_NEGOTIATE_56 and NTLMSSP_NEGOTIATE_128 will both be returned to the client. Clients and servers that set NTLMSSP_NEGOTIATE_SEAL SHOULD set NTLMSSP_NEGOTIATE_128 if it is supported. An alternate name for this field is NTLMSSP_NEGOTIATE_128.<23>',
	NegotiateFlags.r1  : 'This bit is unused and MUST be zero.',
	NegotiateFlags.r2  : 'This bit is unused and MUST be zero.',
	NegotiateFlags.r3  : 'This bit is unused and MUST be zero.',
	NegotiateFlags.NEGOTIATE_VERSION   : 'requests the protocol version number. The data corresponding to this flag is provided in the Version field of the NEGOTIATE_MESSAGE, the CHALLENGE_MESSAGE, and the AUTHENTICATE_MESSAGE.<24> An alternate name for this field is NTLMSSP_NEGOTIATE_VERSION.',
	NegotiateFlags.r4  : 'This bit is unused and MUST be zero.',
	NegotiateFlags.NEGOTIATE_TARGET_INFO   : 'indicates that the TargetInfo fields in the CHALLENGE_MESSAGE (section 2.2.1.2) are populated. An alternate name for this field is NTLMSSP_NEGOTIATE_TARGET_INFO.',
	NegotiateFlags.REQUEST_NON_NT_SESSION_KEY   : ' requests the usage of the LMOWF. An alternate name for this field is NTLMSSP_REQUEST_NON_NT_SESSION_KEY.',
	NegotiateFlags.r5  : 'This bit is unused and MUST be zero.',
	NegotiateFlags.NEGOTIATE_IDENTIFY   : 'requests an identify level token. An alternate name for this field is NTLMSSP_NEGOTIATE_IDENTIFY.',
	NegotiateFlags.NEGOTIATE_EXTENDED_SESSIONSECURITY  : 'requests usage of the NTLM v2 session security. NTLM v2 session security is a misnomer because it is not NTLM v2. It is NTLM v1 using the extended session security that is also in NTLM v2. NTLMSSP_NEGOTIATE_LM_KEY and NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY are mutually exclusive. If both NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY and NTLMSSP_NEGOTIATE_LM_KEY are requested, NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY alone MUST be returned to the client. NTLM v2 authentication session key generation MUST be supported by both the client and the DC in order to be used, and extended  session security signing and sealing requires support from the client and the server in order to be used.<25> An alternate name for this field is NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY.',
	NegotiateFlags.r6  : 'This bit is unused and MUST be zero.',
	NegotiateFlags.TARGET_TYPE_SERVER   : 'TargetName MUST be a server name. The data corresponding to this flag is provided by the server in the TargetName field of the CHALLENGE_MESSAGE. If this bit is set, then NTLMSSP_TARGET_TYPE_DOMAIN MUST NOT be set. This flag MUST be ignored in the NEGOTIATE_MESSAGE and the AUTHENTICATE_MESSAGE. An alternate name for this field is NTLMSSP_TARGET_TYPE_SERVER.',
	NegotiateFlags.TARGET_TYPE_DOMAIN   : 'TargetName MUST be a domain name. The data corresponding to this flag is provided by the server in the TargetName field of the CHALLENGE_MESSAGE. then NTLMSSP_TARGET_TYPE_SERVER MUST NOT be set. This flag MUST be ignored in the NEGOTIATE_MESSAGE and the AUTHENTICATE_MESSAGE. An alternate name for this field is NTLMSSP_TARGET_TYPE_DOMAIN.',
	NegotiateFlags.NEGOTIATE_ALWAYS_SIGN   : ' requests the presence of a signature block on all messages. NTLMSSP_NEGOTIATE_ALWAYS_SIGN MUST be set in the NEGOTIATE_MESSAGE to the server and the CHALLENGE_MESSAGE to the client. NTLMSSP_NEGOTIATE_ALWAYS_SIGN is overridden by NTLMSSP_NEGOTIATE_SIGN and NTLMSSP_NEGOTIATE_SEAL, if they are supported. An alternate name for this field is NTLMSSP_NEGOTIATE_ALWAYS_SIGN.',
	NegotiateFlags.r7  : 'This bit is unused and MUST be zero.',
	NegotiateFlags.NEGOTIATE_OEM_WORKSTATION_SUPPLIED   : 'This flag indicates whether the Workstation field is present. If this flag is not set, the Workstation field MUST be ignored. If this flag is set, the length of the Workstation field specifies whether the workstation name is nonempty or not.<26> An alternate name for this field is NTLMSSP_NEGOTIATE_OEM_WORKSTATION_SUPPLIED.',
	NegotiateFlags.NEGOTIATE_OEM_DOMAIN_SUPPLIED   : 'the domain name is provided (section 2.2.1.1).<27> An alternate name for this field is NTLMSSP_NEGOTIATE_OEM_DOMAIN_SUPPLIED.',
	NegotiateFlags.J   : 'the connection SHOULD be anonymous.<28>',
	NegotiateFlags.r8  : 'This bit is unused and MUST be zero.',
	NegotiateFlags.NEGOTIATE_NTLM   : 'requests usage of the NTLM v1 session security protocol. NTLMSSP_NEGOTIATE_NTLM MUST be set in the NEGOTIATE_MESSAGE to the server and the CHALLENGE_MESSAGE to the client. An alternate name for this field is NTLMSSP_NEGOTIATE_NTLM.',
	NegotiateFlags.r9  : 'This bit is unused and MUST be zero.',
	NegotiateFlags.NEGOTIATE_LM_KEY   : 'requests LAN Manager (LM) session key computation. NTLMSSP_NEGOTIATE_LM_KEY and NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY are mutually exclusive. If both NTLMSSP_NEGOTIATE_LM_KEY and NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY are requested, NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY alone MUST be returned to the client. NTLM v2 authentication session key generation MUST be supported by both the client and the DC in order to be used, and extended session security signing and sealing requires support from the client and the server to be used. An alternate name for this field is NTLMSSP_NEGOTIATE_LM_KEY.',
	NegotiateFlags.NEGOTIATE_DATAGRAM   : 'requests connectionless authentication. If NTLMSSP_NEGOTIATE_DATAGRAM is set, then NTLMSSP_NEGOTIATE_KEY_EXCH MUST always be set in the AUTHENTICATE_MESSAGE to the server and the CHALLENGE_MESSAGE to the client. An alternate name for this field is NTLMSSP_NEGOTIATE_DATAGRAM',
	NegotiateFlags.NEGOTIATE_SEAL   : 'requests session key negotiation for message confidentiality. If the client sends NTLMSSP_NEGOTIATE_SEAL to the server in the NEGOTIATE_MESSAGE, the server MUST return NTLMSSP_NEGOTIATE_SEAL to the client in the CHALLENGE_MESSAGE. Clients and servers that set NTLMSSP_NEGOTIATE_SEAL SHOULD always set NTLMSSP_NEGOTIATE_56 and NTLMSSP_NEGOTIATE_128, if they are supported. An alternate name for this field is NTLMSSP_NEGOTIATE_SEAL.',
	NegotiateFlags.NEGOTIATE_SIGN   : 'requests session key negotiation for message signatures. If the client sends NTLMSSP_NEGOTIATE_SIGN to the server in the NEGOTIATE_MESSAGE, the server MUST return NTLMSSP_NEGOTIATE_SIGN to the client in the CHALLENGE_MESSAGE. An alternate name for this field is NTLMSSP_NEGOTIATE_SIGN.',
	NegotiateFlags.r10 : 'This bit is unused and MUST be zero.',
	NegotiateFlags.REQUEST_TARGET   : 'TargetName field of the CHALLENGE_MESSAGE (section 2.2.1.2) MUST be supplied. An alternate name for this field is NTLMSSP_REQUEST_TARGET.',
	NegotiateFlags.NTLM_NEGOTIATE_OEM   : 'requests OEM character set encoding. An alternate name for this field is NTLM_NEGOTIATE_OEM. See bit A for details.',
	NegotiateFlags.NEGOTIATE_UNICODE   : 'requests Unicode character set encoding. An alternate name for this field is NTLMSSP_NEGOTIATE_UNICODE.',

}


import io

class NTLMSSP_MESSAGE_SIGNATURE_NOEXT:
	def __init__(self):
		self.Version = 1
		self.RandomPad = None
		self.Checksum = None
		self.SeqNum = None

	def to_bytes(self):
		t = self.Version.to_bytes(4, byteorder = 'little', signed = False)
		t += self.RandomPad
		t += self.Checksum
		t += self.SeqNum.to_bytes(4, byteorder = 'little', signed = False)
		return t

	@staticmethod
	def from_bytes(bbuff):
		return NTLMSSP_MESSAGE_SIGNATURE_NOEXT.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		v = NTLMSSP_MESSAGE_SIGNATURE_NOEXT()
		v.Version = int.from_bytes(buff.read(4), byteorder = 'little', signed = False)
		v.RandomPad = buff.read(4)
		v.Checksum = buff.read(4)
		v.SeqNum = int.from_bytes(buff.read(4), byteorder = 'little', signed = False)
			
		return v

	def __repr__(self):
		t  = '== NTLMSSP_MESSAGE_SIGNATURE_NOEXT ==\r\n'
		t += 'Version  : %s\r\n' % self.Version
		t += 'RandomPad  : %s\r\n' % self.RandomPad
		t += 'Checksum  : %s\r\n' % self.Checksum
		t += 'SeqNum  : %s\r\n' % self.SeqNum
		return t